#include <stdio.h>
#include <stdlib.h>

#define  MACHINE_MAIN
#include "comm.h"

int *set_task_list(int start_task_id, int end_task_id, int step,
                   int number_of_tasks)
{
   int *tasks_list;
   int  id;
   int  i;

   if ((tasks_list = (int *) malloc(number_of_tasks * sizeof(int))) == NULL)
   {
      printf("set_task_list(): Out of memory!\n");
      exit(1);
   }

   for (i = number_of_tasks - 1, id = end_task_id; i >= 0; i--, id -= step)
      tasks_list[i] = id;

#ifdef DEBUG_set_task_list
   printf("start id = %d, end id = %d, number_of_tasks = %d\n",
          id + step , end_task_id, number_of_tasks);
#endif

   return(tasks_list);
}

main(int argc, char *argv[])
{
   int i;
   int index;

   MPI_Init(&argc, &argv);

   MPI_Comm_size(MPI_COMM_WORLD, &(machine_info.global.number_of_tasks));
   MPI_Comm_rank(MPI_COMM_WORLD, &(machine_info.global.my_address));
   MPI_Comm_group(MPI_COMM_WORLD, &(machine_info.global.group));
   machine_info.global.comm = MPI_COMM_WORLD;

   machine_info.global.tasks_list
      = set_task_list(0, machine_info.global.number_of_tasks - 1, 1,
                      machine_info.global.number_of_tasks);

   machine_info.grid.dimension = 2;
   
   if ((machine_info.grid.dimension_rank
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   if ((machine_info.grid.periods
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   if ((machine_info.grid.Position
        = (int *) malloc(machine_info.grid.dimension * sizeof(int))) == NULL)
   {
      perror("main()");
      exit(1);
   }

   for (i = machine_info.grid.dimension - 1; i >= 0; i--)
   {
      machine_info.grid.dimension_rank[i] = 0;
      machine_info.grid.periods[i] = 1;
      machine_info.grid.Position[i] = 0;
   }

   MPI_Dims_create(machine_info.global.number_of_tasks,
                   machine_info.grid.dimension,
                   machine_info.grid.dimension_rank);
   MPI_Cart_create(MPI_COMM_WORLD, machine_info.grid.dimension,
                   machine_info.grid.dimension_rank,
                   machine_info.grid.periods, 1,
                   &(machine_info.grid.comm));
   MPI_Comm_group(machine_info.grid.comm, &(machine_info.grid.group));
   MPI_Comm_size(machine_info.grid.comm,
                 &(machine_info.grid.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.comm, &(machine_info.grid.my_address));
   MPI_Cart_coords(machine_info.grid.comm, machine_info.grid.my_address,
                   machine_info.grid.dimension, machine_info.grid.Position);

   /* Building rows */
   MPI_Comm_split(machine_info.grid.comm, machine_info.grid.Position[0],
                  0, &(machine_info.grid.row.comm));
   MPI_Comm_group(machine_info.grid.row.comm,
                  &(machine_info.grid.row.group));
   MPI_Comm_group(machine_info.grid.row.comm, &(machine_info.grid.row.group));
   MPI_Comm_size(machine_info.grid.row.comm,
                 &(machine_info.grid.row.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.row.comm,
                 &(machine_info.grid.row.my_address));

   /* Building columns */
   MPI_Comm_split(machine_info.grid.comm, machine_info.grid.Position[1],
                  0, &(machine_info.grid.col.comm));
   MPI_Comm_group(machine_info.grid.col.comm,
                  &(machine_info.grid.col.group));
   MPI_Comm_group(machine_info.grid.col.comm, &(machine_info.grid.col.group));
   MPI_Comm_size(machine_info.grid.col.comm,
                 &(machine_info.grid.col.number_of_tasks));
   MPI_Comm_rank(machine_info.grid.col.comm,
                 &(machine_info.grid.col.my_address));

#if 0
   MPE_Init_log();
   if (machine_info.my_address == 0)
   {
      MPE_Describe_state(1, 2, "Send msg", "blue:vlines3");
      MPE_Describe_state(3, 4, "Receive msg", "red:gray3");
      MPE_Describe_state(5, 6, "Barrier", "green:light_gray");
      MPE_Describe_state(7, 8, "Async. Send msg", "AliceBlue");
      MPE_Describe_state(9, 10, "Async. Receive msg", "LightBlue");
      MPE_Describe_state(11, 12, "Async. wait", "IndianRed");
   }
   MPE_Start_log();
#endif
 
   printf("node(");
   for (i = 0; i < machine_info.grid.dimension - 1; i++)
   {
      printf("%d, ", machine_info.grid.Position[i]); 
   }
   printf("%d): hello\n", machine_info.grid.Position[i]);

   index = machine_info.grid.Position[0];
   MPI_Bcast(&index, 1, MPI_INT, 0, machine_info.grid.row.comm);

   printf("MPI_Bcast receive %d - %d\n", index, machine_info.grid.Position[0]);

#if 0
   MPE_Stop_log();
   MPE_Finish_log("lu.log");
#endif

   MPI_Finalize();
   exit(0);
}

